import asyncio
import base64
import json
import os
import websockets
import click


CHUNK_SIZE = 5 * 1024 * 1024  # 5 MiB


target: str = None
port_range: list[int] = None
ip_list: list[str] = None


# Print iterations progress
def printProgressBar(
    iteration,
    total,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="█",
    printEnd="\r",
):
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)
    # Print New Line on Complete
    if iteration == total:
        print()


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


async def stream_receive():
    global target, scrt, ip_list, port_range
    welcome = None
    for ip in ip_list:
        for port in range(port_range[0], port_range[1] + 1):
            uri = f"ws://{ip}:{port}"
            try:
                async with websockets.connect(
                    uri,
                    max_size=int(
                        CHUNK_SIZE * 1.25
                    ),  # Allow some overhead for encoding and headers
                    ping_interval=60,
                    ping_timeout=None,
                    additional_headers={"Authorization": f"Bearer {scrt}"},
                ) as websocket:
                    with open(target, "wb") as f:
                        chunk_count = 0
                        async for message in websocket:
                            if welcome is None:
                                welcome = json.loads(message.decode("utf-8"))
                                print(
                                    f"Transfering {sizeof_fmt(welcome['file_size'])}..."
                                )
                                continue
                            if message == "EOF":
                                break
                            f.write(message)
                            chunk_count += 1
                            printProgressBar(
                                chunk_count, welcome["num_chunks"], length=40
                            )
                    print("File received successfully.")
                    return
            except (
                OSError,
                websockets.exceptions.InvalidStatus,
                websockets.exceptions.InvalidMessage,
            ):
                continue
    print("Unable to establish connection to any provided IPs/ports.")


async def stream_send():
    global target, scrt, ip_list, port_range
    for ip in ip_list:
        for port in range(port_range[0], port_range[1] + 1):
            uri = f"ws://{ip}:{port}"
            try:
                async with websockets.connect(
                    uri,
                    max_size=int(
                        CHUNK_SIZE * 1.25
                    ),  # Allow some overhead for encoding and headers
                    ping_interval=60,
                    ping_timeout=None,
                    additional_headers={"Authorization": f"Bearer {scrt}"},
                ) as websocket:
                    file_size = os.stat(target).st_size
                    num_chunks = (file_size + CHUNK_SIZE - 1) // CHUNK_SIZE
                    print(f"Transfering {sizeof_fmt(file_size)}...")
                    chunk_count = 0
                    with open(target, "rb") as f:
                        while chunk := f.read(CHUNK_SIZE):
                            await websocket.send(chunk, text=False)
                            chunk_count += 1
                            printProgressBar(chunk_count, num_chunks, length=40)
                    await websocket.send("EOF")  # Signal end of file
                    print("File sent successfully.")
                    return
            except (
                OSError,
                websockets.exceptions.InvalidStatus,
                websockets.exceptions.InvalidMessage,
            ):
                continue
    print("Unable to establish connection to any provided IPs/ports.")


def set_coordinates(coordinates):
    global scrt, port_range, ip_list
    try:
        json_str = base64.urlsafe_b64decode(coordinates).decode("utf-8")
        data = json.loads(json_str)

        scrt = data["secret"]
        port_range = data["ports"]
        ip_list = data["ips"]
    except (json.JSONDecodeError, KeyError, base64.binascii.Error) as e:
        raise click.ClickException("Invalid coordinates format") from e


@click.command()
@click.option("--path", help="The source path of the file to be sent.", required=True)
@click.option(
    "--coordinates",
    help="Secret coordinates used to establish a connection",
    required=True,
)
def send(path, coordinates):
    global target
    set_coordinates(coordinates)
    target = path
    asyncio.run(stream_send())


@click.command()
@click.option(
    "--coordinates",
    help="Secret coordinates used to establish a connection",
    required=True,
)
@click.option("--path", help="The target path of the incoming file.", required=True)
def receive(path, coordinates):
    global operation, target
    set_coordinates(coordinates)
    target = path
    asyncio.run(stream_receive())



import click
from streamer.streamer_client import send, receive
from streamer.streamer_server import server
from importlib.metadata import version


def get_version():
    try:
        return version("firecrest-streamer")
    except Exception:
        return "dev"


@click.group(invoke_without_command=True)
@click.option("--version", is_flag=True, help="Show version and exit.")
@click.pass_context
def cli(ctx, version):
    if version:
        click.echo(f"Firecrest streamer version: {get_version()}")
        ctx.exit()


# Register groups
cli.add_command(send)
cli.add_command(receive)
cli.add_command(server)


if __name__ == "__main__":
    cli()



import asyncio
import base64
from enum import Enum
import http
import json
import os
import signal
import websockets
from websockets.asyncio.server import serve
import click


CHUNK_SIZE = 5 * 1024 * 1024  # 5 MiB


class Operation(Enum):
    send = "send"
    receive = "receive"


operation: Operation = None
target: str = None
secret: str = None
port_range: tuple[int, int] = None
ips: list[str] = None
host: str = None
wait_timeout: int = None
inbound_transfer_limit: int = None
timeout_handle: asyncio.Handle = None


async def stream_receive(websocket):
    global operation, target, inbound_transfer_limit
    print("Client connected.")
    transfer_size = 0
    with open(target, "wb") as f:
        try:
            async for message in websocket:
                if message == "EOF":
                    print("File transfer complete.")
                    break
                transfer_size += CHUNK_SIZE
                if transfer_size > inbound_transfer_limit:
                    print(
                        "Inbound transfer limit exceeded, max allowed transfer size: {inbound_transfer_limit} bytes Aborting transfer."
                    )
                    await websocket.close(
                        code=1009,
                        reason=f"Inbound transfer limit exceeded, max allowed transfer size: {inbound_transfer_limit} bytes.",
                    )
                    os.remove(target)
                    websocket.server.close()
                    return
                f.write(message)
        except websockets.ConnectionClosed:
            print("Connection closed unexpectedly.")
    print(f"File {target} received successfully.")
    websocket.server.close()


async def stream_send(websocket):
    global operation, target
    print("Client connected.")
    file_size = os.stat(target).st_size
    num_chunks = (file_size + CHUNK_SIZE - 1) // CHUNK_SIZE
    await websocket.send(
        json.dumps({"num_chunks": num_chunks, "file_size": file_size}).encode(
            encoding="utf-8"
        )
    )
    with open(target, "rb") as f:
        try:
            while True:
                chunk = f.read(CHUNK_SIZE)
                if not chunk:
                    break
                await websocket.send(chunk, text=False)
            await websocket.send("EOF")  # signal end of file
        except websockets.ConnectionClosed:
            print("Connection closed unexpectedly.")
    print(f"File {target} sent successfully.")
    websocket.server.close()


def process_request(connection, request):
    global secret, timeout_handle
    if "Authorization" not in request.headers:
        return connection.respond(
            http.HTTPStatus.UNAUTHORIZED, "Missing Authorization header\n"
        )

    authorization = request.headers["Authorization"]
    if authorization is None:
        return connection.respond(http.HTTPStatus.UNAUTHORIZED, "Missing token\n")

    token = authorization.split("Bearer ")[-1]
    if token is None or token != secret:
        return connection.respond(http.HTTPStatus.FORBIDDEN, "Invalid secret\n")

    timeout_handle.cancel()


async def stream():
    global secret, port_range, ips, host, wait_timeout, timeout_handle
    start_port, end_port = port_range
    for port in range(start_port, end_port + 1):
        try:
            async with serve(
                stream_receive if operation == Operation.receive else stream_send,
                host,
                port,
                max_size=int(
                    CHUNK_SIZE * 1.25
                ),  # Allow some overhead for encoding and headers
                ping_interval=60,
                ping_timeout=None,
                process_request=process_request,
            ) as server:
                print(f"Server is listening on ws://{host}:{port}")
                coordinates = {
                    "ports": [start_port, end_port],
                    "ips": ips,
                    "secret": secret,
                }
                encoded = base64.urlsafe_b64encode(
                    json.dumps(coordinates).encode("utf-8")
                ).decode("utf-8")

                print(f"Use these coordinates to connect: {encoded}", flush=True)

                loop = asyncio.get_running_loop()
                loop.add_signal_handler(signal.SIGTERM, server.close)
                timeout_handle = loop.call_later(wait_timeout, server.close)
                await server.wait_closed()
            break
        except OSError:
            print(f"Server unable to bing on port: {port}")
            continue


@click.group()
@click.option(
    "--secret",
    "_secret",
    help="A shared secret required to initiate the transfer",
    required=True,
)
@click.option(
    "--public-ips",
    "_ips",
    help="A list of public IPs where the streamer server might run.",
    default=["localhost"],
    multiple=True,
)
@click.option(
    "--host",
    "_host",
    help="The interface to use for listening incoming connections",
    default="localhost",
)
@click.option(
    "--port-range",
    "_port_range",
    type=(int, int),
    help="A range of ports to pick from to listen for incoming connections e.g. --port-range 5665 5670",
    default=(5665, 5670),
)
@click.option(
    "--wait-timeout",
    "_wait_timeout",
    help="How long to wait for a connection before exiting (in seconds)",
    default=60 * 60 * 24,  # 24h
)
@click.option(
    "--inbound-transfer-limit",
    "_inbound_transfer_limit",
    help="Limit how much data can be received (in bytes)",
    default=5 * 1024 * 1024 * 1024,  # 5GB
)
def server(_secret, _ips, _host, _port_range, _wait_timeout, _inbound_transfer_limit):
    global secret, port_range, ips, wait_timeout, inbound_transfer_limit, host
    secret = _secret
    port_range = _port_range
    ips = _ips
    host = _host
    wait_timeout = _wait_timeout
    inbound_transfer_limit = _inbound_transfer_limit


@server.command()
@click.option("--path", help="The target path of the file to be sent.", required=True)
def send(path):
    global operation, target
    operation = Operation.send
    target = path
    asyncio.run(stream())


@server.command()
@click.option(
    "--path", help="The target path of the file to be received.", required=True
)
def receive(path):
    global operation, target
    operation = Operation.receive
    target = path
    asyncio.run(stream())
